@@ -1,7 +1,7 @@
 from enum import Enum
 
-from biolib.typing_utils import TypedDict, List, Optional, Dict, Literal
 from biolib.biolib_api_client.common_types import SemanticVersion
+from biolib.typing_utils import Dict, List, Literal, Optional, TypedDict
 
 
 class AppVersionSlim(SemanticVersion):
@@ -16,6 +16,7 @@ class AppVersion(AppVersionSlim):
     source_code_license: str
     stdout_render_type: Literal['text', 'markdown']
     main_output_file: Optional[str]
+    app_uri: str
 
 
 class App(TypedDict):
@@ -31,6 +32,7 @@ class App(TypedDict):
     public_id: str
     state: str
     resource_uri: str
+    type: str
 
 
 class AppGetResponse(TypedDict):
@@ -70,6 +72,7 @@ class _Module(TypedDict):
     command: str
     environment: Literal['biolib-app', 'biolib-custom', 'biolib-ecr']
     image_uri: str
+    absolute_image_uri: str
     estimated_image_size_bytes: Optional[int]
     input_files_mappings: List[FilesMapping]
     large_file_systems: List[LargeFileSystemMapping]
