@@ -1,4 +1,11 @@
 from biolib.biolib_binary_format.base_bbf_package import BioLibBinaryFormatBasePackage
+from biolib.typing_utils import TypedDict, Dict, List
+
+
+class ModuleInputDict(TypedDict):
+    stdin: bytes
+    files: Dict[str, bytes]
+    arguments: List[str]
 
 
 class ModuleInput(BioLibBinaryFormatBasePackage):
@@ -36,7 +43,7 @@ class ModuleInput(BioLibBinaryFormatBasePackage):
 
         return bbf_data
 
-    def deserialize(self):
+    def deserialize(self) -> ModuleInputDict:
         version = self.get_data(1, output_type='int')
         package_type = self.get_data(1, output_type='int')
         self.check_version_and_type(version=version, package_type=package_type, expected_package_type=self.package_type)
@@ -62,4 +69,4 @@ class ModuleInput(BioLibBinaryFormatBasePackage):
             data = self.get_data(data_len)
             files[path] = bytes(data)
 
-        return {'stdin': stdin, 'arguments': arguments, 'files': files}
+        return ModuleInputDict(stdin=stdin, arguments=arguments, files=files)
