@@ -25,11 +25,6 @@ class LfsCacheStateDict(TypedDict):
     large_file_systems: Dict[UuidStr, LargeFileSystemCache]
 
 
-class DockerAuthConfig(TypedDict):
-    username: str
-    password: str
-
-
 class DockerImageInfo(TypedDict):
     last_used_at: str
     estimated_image_size_bytes: int
