@@ -12,9 +12,9 @@ import os
 import signal
 from types import FrameType
 
-import requests
 from docker.models.networks import Network  # type: ignore
 
+from biolib._internal.http_client import HttpClient
 from biolib.biolib_binary_format.stdout_and_stderr import StdoutAndStderr
 from biolib.compute_node.job_worker.job_legacy_input_wait_timeout_thread import JobLegacyInputWaitTimeout
 from biolib.compute_node.job_worker.job_storage import JobStorage
@@ -97,7 +97,11 @@ class JobWorker:
         if socket_port:
             self._connect_to_parent()
 
-    def _handle_exit_gracefully(self, signum: int, frame: FrameType) -> None:  # pylint: disable=unused-argument
+    def _handle_exit_gracefully(
+            self,
+            signum: int,
+            frame: Optional[FrameType],  # pylint: disable=unused-argument
+    ) -> None:
         job_id = self._root_job_wrapper["job"]["public_id"] if self._root_job_wrapper else None
         logger_no_user_data.debug(
             f'_JobWorker ({job_id}) got exit signal {signal.Signals(signum).name}'  # pylint: disable=no-member
@@ -129,7 +133,8 @@ class JobWorker:
                         ).start()
 
                         try:
-                            module_input_serialized = JobStorage.get_module_input(job=job)
+                            module_input_path = os.path.join(self.job_temporary_dir, JobStorage.module_input_file_name)
+                            JobStorage.download_module_input(job=job, path=module_input_path)
                         except StorageDownloadFailed:
                             # Expect module input to be handled in a separate ModuleInput package
                             self._legacy_input_wait_timeout_thread = JobLegacyInputWaitTimeout(
@@ -143,7 +148,7 @@ class JobWorker:
                             raise error
 
                         try:
-                            self._run_root_job(module_input_serialized)
+                            self._run_root_job(module_input_path)
 
                         # This error occurs when trying to access the container after the job worker has cleaned it up.
                         # In that case stop the computation.
@@ -161,7 +166,9 @@ class JobWorker:
                         self._legacy_input_wait_timeout_thread.stop()
 
                     try:
-                        self._run_root_job(package)
+                        module_input_path = os.path.join(self.job_temporary_dir, JobStorage.module_input_file_name)
+                        open(module_input_path, 'wb').write(package)
+                        self._run_root_job(module_input_path)
 
                     # This error occurs when trying to access the container after the job worker has cleaned it up.
                     # In that case stop the computation.
@@ -188,11 +195,13 @@ class JobWorker:
     def _cleanup(self) -> None:
         self.is_cleaning_up = True
 
+        logger_no_user_data.debug('Cleaning up executers...')
         for executor in self._executors:
             executor.cleanup()
 
         proxy_count = len(self._remote_host_proxies)
         if proxy_count > 0:
+            logger_no_user_data.debug('Cleaning up proxies...')
             proxy_cleanup_start_time = time()
 
             for proxy in self._remote_host_proxies:
@@ -205,10 +214,12 @@ class JobWorker:
             self._remote_host_proxies = []
             logger_no_user_data.debug(f'Cleaned up {proxy_count} proxies in {time() - proxy_cleanup_start_time}')
 
+        logger_no_user_data.debug('Cleaning up networks...')
         self._cleanup_network(self._internal_network)
         self._internal_network = None
         self._cleanup_network(self._public_network)
         self._public_network = None
+        logger_no_user_data.debug('Cleaned up networks...')
 
     @staticmethod
     def _cleanup_network(network: Optional[Network]) -> None:
@@ -307,7 +318,7 @@ class JobWorker:
                         self._public_network,
                         self._internal_network,
                         job_id,
-                        ports
+                        ports,
                     )
                     remote_host_proxy.start()
                     self._remote_host_proxies.append(remote_host_proxy)
@@ -325,15 +336,15 @@ class JobWorker:
     def _run_app_version(
             self,
             app_version_id: str,
-            module_input_serialized: bytes,
+            module_input_path: str,
             caller_job: CreatedJobDict,
             main_module_output_path: str,
     ) -> None:
         job: CreatedJobDict = BiolibJobApi.create(app_version_id, caller_job=caller_job['public_id'])
         self._jobs[job['public_id']] = job
-        self._run_job(job, module_input_serialized, main_module_output_path)
+        self._run_job(job, module_input_path, main_module_output_path)
 
-    def _run_job(self, job: CreatedJobDict, module_input_serialized: bytes, main_module_output_path: str) -> None:
+    def _run_job(self, job: CreatedJobDict, module_input_path: str, main_module_output_path: str) -> None:
         job_uuid = job['public_id']
         logger_no_user_data.info(f'Job "{job_uuid}" running...')
         if self._root_job_wrapper is None:
@@ -393,15 +404,15 @@ class JobWorker:
                 cloud_job=self._root_job_wrapper['cloud_job'],
                 large_file_systems=lfs_dict,
                 module=main_module,
+                module_input_path=module_input_path,
+                module_output_path=main_module_output_path,
                 remote_host_proxies=self._remote_host_proxies,
                 root_job_id=root_job_id,
                 runtime_zip_bytes=runtime_zip_bytes,
                 send_status_update=self._send_status_update,
                 send_system_exception=self.send_system_exception,
                 send_stdout_and_stderr=self.send_stdout_and_stderr,
-            ),
-            module_input_serialized,
-            main_module_output_path,
+            )
         )
 
         if utils.IS_RUNNING_IN_CLOUD:
@@ -411,16 +422,20 @@ class JobWorker:
     def _run_module(
             self,
             options: LocalExecutorOptions,
-            module_input_serialized: bytes,
-            module_output_path: str,
     ) -> None:
         module = options['module']
         job_id = options['job']['public_id']
+        module_output_path = options['module_output_path']
+        module_input_path = options['module_input_path']
         logger_no_user_data.debug(f'Job "{job_id}" running module "{module["name"]}"...')
-        executor_instance: DockerExecutor
 
+        executor_instance: DockerExecutor
         if module['environment'] == ModuleEnvironment.BIOLIB_APP.value:
+            if not self.job_temporary_dir:
+                raise BioLibError('Undefined job_temporary_dir')
             logger_no_user_data.debug(f'Job "{job_id}" starting child job...')
+            with open(module_input_path,'rb') as fp:
+                module_input_serialized = fp.read()
             module_input = ModuleInput(module_input_serialized).deserialize()
             module_input_with_runtime_zip = self._add_runtime_zip_and_command_to_module_input(options, module_input)
             module_input_with_runtime_zip_serialized = ModuleInput().serialize(
@@ -428,9 +443,11 @@ class JobWorker:
                 arguments=module_input_with_runtime_zip['arguments'],
                 files=module_input_with_runtime_zip['files'],
             )
+            module_input_path_new = os.path.join(self.job_temporary_dir, "runtime." + JobStorage.module_input_file_name)
+            open(module_input_path_new, 'wb').write(module_input_with_runtime_zip_serialized)
             return self._run_app_version(
                 module['image_uri'],
-                module_input_with_runtime_zip_serialized,
+                module_input_path_new,
                 options['job'],
                 module_output_path,
             )
@@ -456,7 +473,7 @@ class JobWorker:
             # Log memory and disk before pulling and executing module
             log_disk_and_memory_usage_info()
 
-        executor_instance.execute_module(module_input_serialized, module_output_path)
+        executor_instance.execute_module()
 
     def _connect_to_parent(self):
         try:
@@ -522,27 +539,22 @@ class JobWorker:
         runtime_zip_bytes: Optional[bytes] = self._app_version_id_to_runtime_zip.get(app_version['public_id'])
 
         if runtime_zip_bytes is None:
-            if root_job_id == utils.RUN_DEV_JOB_ID:
-                with open(runtime_zip_url, mode='rb') as runtime_zip_file:
-                    runtime_zip_bytes = runtime_zip_file.read()
+            self._send_status_update(StatusUpdate(progress=25, log_message='Downloading Source Files...'))
 
-            else:
-                self._send_status_update(StatusUpdate(progress=25, log_message='Downloading Source Files...'))
-
-                start_time = time()
-                logger_no_user_data.debug(f'Job "{root_job_id}" downloading runtime zip...')
-                try:
-                    runtime_zip_bytes = requests.get(runtime_zip_url).content
-                except Exception as exception:
-                    raise ComputeProcessException(
-                        exception,
-                        SystemExceptionCodes.FAILED_TO_DOWNLOAD_RUNTIME_ZIP.value,
-                        self.send_system_exception,
-                        may_contain_user_data=False
-                    ) from exception
-                finally:
-                    download_time = time() - start_time
-                    logger_no_user_data.debug(f'Job "{root_job_id}" download of runtime zip took: {download_time}s')
+            start_time = time()
+            logger_no_user_data.debug(f'Job "{root_job_id}" downloading runtime zip...')
+            try:
+                runtime_zip_bytes = HttpClient.request(url=runtime_zip_url).content
+            except Exception as exception:
+                raise ComputeProcessException(
+                    exception,
+                    SystemExceptionCodes.FAILED_TO_DOWNLOAD_RUNTIME_ZIP.value,
+                    self.send_system_exception,
+                    may_contain_user_data=False
+                ) from exception
+            finally:
+                download_time = time() - start_time
+                logger_no_user_data.debug(f'Job "{root_job_id}" download of runtime zip took: {download_time}s')
 
             self._app_version_id_to_runtime_zip[app_version['public_id']] = runtime_zip_bytes
 
@@ -587,7 +599,7 @@ class JobWorker:
                 may_contain_user_data=False
             ) from exception
 
-    def _run_root_job(self, module_input_serialized: bytes) -> str:
+    def _run_root_job(self, module_input_path: str) -> str:
         # Make typechecker happy
         if not self._root_job_wrapper or not self.job_temporary_dir:
             raise BioLibError('Undefined job_wrapper or job_temporary_dir')
@@ -595,7 +607,7 @@ class JobWorker:
         main_module_output_path = os.path.join(self.job_temporary_dir, JobStorage.module_output_file_name)
         self._run_job(
             job=self._root_job_wrapper['job'],
-            module_input_serialized=module_input_serialized,
+            module_input_path=module_input_path,
             main_module_output_path=main_module_output_path,
         )
         self._send_status_update(StatusUpdate(progress=94, log_message='Computation finished'))
@@ -614,7 +626,9 @@ class JobWorker:
                     job_temporary_dir=job_temporary_dir,
                 )
                 self._start_network_and_remote_host_proxies(job_dict)
-                module_output_path = self._run_root_job(module_input_serialized)
+                module_input_path = os.path.join(self.job_temporary_dir, JobStorage.module_input_file_name)
+                open(module_input_path, 'wb').write(module_input_serialized)
+                module_output_path = self._run_root_job(module_input_path)
                 with open(module_output_path, mode='rb') as module_output_file:
                     module_output_serialized = module_output_file.read()
                 return ModuleOutputV2(InMemoryIndexableBuffer(module_output_serialized))
