@@ -1,21 +1,22 @@
+import base64
 import io
-import tarfile
 import subprocess
+import tarfile
 import time
+from urllib.parse import urlparse
 
-from docker.models.containers import Container  # type: ignore
 from docker.errors import ImageNotFound  # type: ignore
+from docker.models.containers import Container  # type: ignore
 from docker.models.images import Image  # type: ignore
 from docker.models.networks import Network  # type: ignore
 
 from biolib import utils
-from biolib.biolib_errors import BioLibError
-from biolib.compute_node.cloud_utils import CloudUtils
-from biolib.typing_utils import Optional, List
-from biolib.biolib_api_client import RemoteHost
+from biolib.biolib_api_client import BiolibApiClient, RemoteHost
 from biolib.biolib_docker_client import BiolibDockerClient
-from biolib.biolib_logging import logger, logger_no_user_data
-from biolib.biolib_api_client import BiolibApiClient
+from biolib.biolib_errors import BioLibError
+from biolib.biolib_logging import logger_no_user_data
+from biolib.compute_node.cloud_utils import CloudUtils
+from biolib.typing_utils import List, Optional
 
 
 # Prepare for remote hosts with specified port
@@ -24,34 +25,27 @@ class RemoteHostExtended(RemoteHost):
 
 
 class RemoteHostProxy:
-    _DOCKER_IMAGE_URI = 'public.ecr.aws/h5y4b3l1/biolib-remote-host-proxy:latest'
-    _TRAFFIC_FORWARDER_PORT_OFFSET = 10000  # Port offset relative to port of a VSOCK proxy
-
     def __init__(
-            self,
-            remote_host: RemoteHost,
-            public_network: Network,
-            internal_network: Optional[Network],
-            job_id: str,
-            ports: List[int]
+        self,
+        remote_host: RemoteHost,
+        public_network: Network,
+        internal_network: Optional[Network],
+        job_id: str,
+        ports: List[int],
     ):
         self.is_app_caller_proxy = remote_host['hostname'] == 'AppCallerProxy'
-
-        # Default to port 443 for now until backend serves remote_hosts with port specified
-        self._remote_host: RemoteHostExtended = RemoteHostExtended(
-            hostname=remote_host['hostname'],
-            ports=ports
-        )
+        self._remote_host: RemoteHostExtended = RemoteHostExtended(hostname=remote_host['hostname'], ports=ports)
         self._public_network: Network = public_network
         self._internal_network: Optional[Network] = internal_network
 
         if not job_id:
             raise Exception('RemoteHostProxy missing argument "job_id"')
 
-        self._name = f"biolib-remote-host-proxy-{job_id}-{self.hostname}"
+        self._name = f'biolib-remote-host-proxy-{job_id}-{self.hostname}'
         self._job_uuid = job_id
         self._container: Optional[Container] = None
         self._enclave_traffic_forwarder_processes: List[subprocess.Popen] = []
+        self._docker = BiolibDockerClient().get_docker_client()
 
     @property
     def hostname(self) -> str:
@@ -81,7 +75,7 @@ class RemoteHostProxy:
             try:
                 self._container = docker.containers.create(
                     detach=True,
-                    image=self._get_nginx_docker_image(),
+                    image=self._get_biolib_remote_host_proxy_image(),
                     name=self._name,
                     network=self._public_network.name,
                 )
@@ -130,45 +124,47 @@ class RemoteHostProxy:
         for process in self._enclave_traffic_forwarder_processes:
             process.terminate()
 
-    def _get_nginx_docker_image(self) -> Image:
-        docker = BiolibDockerClient.get_docker_client()
+    def _get_biolib_remote_host_proxy_image(self) -> Image:
+        if utils.IS_RUNNING_IN_CLOUD:
+            try:
+                logger_no_user_data.debug('Getting local Docker image for remote host proxy')
+                return self._docker.images.get('biolib-remote-host-proxy:latest')
+            except ImageNotFound:
+                logger_no_user_data.debug(
+                    'Local Docker image for remote host proxy not available. Falling back to public image...'
+                )
+
+        public_image_uri = 'public.ecr.aws/h5y4b3l1/biolib-remote-host-proxy:latest'
         try:
-            return docker.images.get(self._DOCKER_IMAGE_URI)
+            logger_no_user_data.debug('Getting public Docker image for remote host proxy')
+            return self._docker.images.get(public_image_uri)
         except ImageNotFound:
-            logger.debug('Pulling remote host docker image...')
-            return docker.images.pull(self._DOCKER_IMAGE_URI)
+            logger_no_user_data.debug('Pulling public Docker image for remote host proxy')
+            return self._docker.images.pull(public_image_uri)
 
     def _write_nginx_config_to_container(self, upstream_server_name: str, upstream_server_ports: List[int]) -> None:
         if not self._container:
             raise Exception('RemoteHostProxy container not defined when attempting to write NGINX config')
 
         docker = BiolibDockerClient.get_docker_client()
-        base_url = BiolibApiClient.get().base_url
+        upstream_hostname = urlparse(BiolibApiClient.get().base_url).hostname
         if self.is_app_caller_proxy:
-            logger_no_user_data.debug(f'Job "{self._job_uuid}" writing config for and starting App Caller Proxy...')
-            if utils.BIOLIB_CLOUD_BASE_URL:
-                cloud_base_url = utils.BIOLIB_CLOUD_BASE_URL
-            else:
-                if base_url in ('https://biolib.com', 'https://staging-elb.biolib.com'):
-                    cloud_base_url = 'https://biolibcloud.com'
-                else:
-                    raise BioLibError('Calling apps inside apps is not supported in local compute environment')
+            if not utils.IS_RUNNING_IN_CLOUD:
+                raise BioLibError('Calling apps inside apps is not supported in local compute environment')
 
-            if utils.IS_RUNNING_IN_CLOUD:
-                config = CloudUtils.get_webserver_config()
-                s3_results_bucket_name = config['s3_general_storage_bucket_name']  # pylint: disable=unsubscriptable-object
-                s3_results_base_url = f'https://{s3_results_bucket_name}.s3.amazonaws.com'
-            else:
-                if base_url in ('https://biolib.com', 'https://staging-elb.biolib.com'):
-                    s3_results_base_url = 'https://biolib-cloud-api.s3.amazonaws.com'
-                else:
-                    raise BioLibError("Calling apps inside apps locally is only supported on biolib.com")
+            logger_no_user_data.debug(f'Job "{self._job_uuid}" writing config for and starting App Caller Proxy...')
+            config = CloudUtils.get_webserver_config()
+            compute_node_uuid = config['compute_node_info']['public_id']
+            compute_node_auth_token = config['compute_node_info']['auth_token']
 
             # TODO: Get access_token from new API class instead
             access_token = BiolibApiClient.get().access_token
             bearer_token = f'Bearer {access_token}' if access_token else ''
 
-            nginx_config = f'''
+            biolib_index_basic_auth = f'compute_node|admin:{compute_node_auth_token},{self._job_uuid}'
+            biolib_index_basic_auth_base64 = base64.b64encode(biolib_index_basic_auth.encode('utf-8')).decode('utf-8')
+
+            nginx_config = f"""
 events {{
   worker_connections  1024;
 }}
@@ -185,11 +181,6 @@ http {{
         default    "";
     }}
 
-    map $request_method $bearer_token_on_patch {{
-        PATCH      "{bearer_token}";
-        default    "";
-    }}
-
     map $request_method $bearer_token_on_patch_and_get {{
         PATCH      "{bearer_token}";
         GET        "{bearer_token}";
@@ -198,10 +189,11 @@ http {{
 
     server {{
         listen       80;
-        resolver 127.0.0.11 valid=30s;
+        resolver 127.0.0.11 ipv6=off valid=30s;
+        set $upstream_hostname {upstream_hostname};
 
         location ~* "^/api/jobs/cloud/(?<job_id>[a-z0-9-]{{36}})/status/$" {{
-            proxy_pass               {base_url}/api/jobs/cloud/$job_id/status/;
+            proxy_pass               https://$upstream_hostname/api/jobs/cloud/$job_id/status/;
             proxy_set_header         authorization $bearer_token_on_get;
             proxy_set_header         cookie "";
             proxy_ssl_server_name    on;
@@ -209,35 +201,35 @@ http {{
 
         location ~* "^/api/jobs/cloud/$" {{
             # Note: Using $1 here as URI part from regex must be used for proxy_pass
-            proxy_pass               {base_url}/api/jobs/cloud/$1;
+            proxy_pass               https://$upstream_hostname/api/jobs/cloud/$1;
             proxy_set_header         authorization $bearer_token_on_post;
             proxy_set_header         cookie "";
             proxy_ssl_server_name    on;
         }}
 
         location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/start_upload/$" {{
-            proxy_pass               {base_url}/api/jobs/$job_id/storage/input/start_upload/;
+            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/storage/input/start_upload/;
             proxy_set_header         authorization "";
             proxy_set_header         cookie "";
             proxy_ssl_server_name    on;
         }}
 
         location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/presigned_upload_url/$" {{
-            proxy_pass               {base_url}/api/jobs/$job_id/storage/input/presigned_upload_url/$is_args$args;
+            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/storage/input/presigned_upload_url/$is_args$args;
             proxy_set_header         authorization "";
             proxy_set_header         cookie "";
             proxy_ssl_server_name    on;
         }}
 
         location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/complete_upload/$" {{
-            proxy_pass               {base_url}/api/jobs/$job_id/storage/input/complete_upload/;
+            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/storage/input/complete_upload/;
             proxy_set_header         authorization "";
             proxy_set_header         cookie "";
             proxy_ssl_server_name    on;
         }}
-        
+
         location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/main_result/$" {{
-            proxy_pass                  {base_url}/api/jobs/$job_id/main_result/;
+            proxy_pass                  https://$upstream_hostname/api/jobs/$job_id/main_result/;
             proxy_set_header            authorization "";
             proxy_set_header            cookie "";
             proxy_pass_request_headers  on;
@@ -245,7 +237,7 @@ http {{
         }}
 
         location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/$" {{
-            proxy_pass               {base_url}/api/jobs/$job_id/;
+            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/;
             proxy_set_header         authorization $bearer_token_on_patch_and_get;
             proxy_set_header         caller-job-uuid "{self._job_uuid}";
             proxy_set_header         cookie "";
@@ -254,7 +246,7 @@ http {{
 
         location ~* "^/api/jobs/create_job_with_data/$" {{
             # Note: Using $1 here as URI part from regex must be used for proxy_pass
-            proxy_pass               {base_url}/api/jobs/create_job_with_data/$1;
+            proxy_pass               https://$upstream_hostname/api/jobs/create_job_with_data/$1;
             proxy_set_header         authorization $bearer_token_on_post;
             proxy_set_header         caller-job-uuid "{self._job_uuid}";
             proxy_set_header         cookie "";
@@ -263,58 +255,118 @@ http {{
 
         location ~* "^/api/jobs/$" {{
             # Note: Using $1 here as URI part from regex must be used for proxy_pass
-            proxy_pass               {base_url}/api/jobs/$1;
+            proxy_pass               https://$upstream_hostname/api/jobs/$1;
             proxy_set_header         authorization $bearer_token_on_post;
             proxy_set_header         caller-job-uuid "{self._job_uuid}";
             proxy_set_header         cookie "";
             proxy_ssl_server_name    on;
         }}
 
+        location ~ "^/api/jobs/{self._job_uuid}/notes/$" {{
+            # Note: Using $1 here as URI part from regex must be used for proxy_pass
+            proxy_pass               https://$upstream_hostname/api/jobs/{self._job_uuid}/notes/$1;
+            proxy_set_header         authorization "";
+            proxy_set_header         job-auth-token "";
+            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
+            proxy_set_header         compute-node-uuid "{compute_node_uuid}";
+            proxy_set_header         cookie "";
+            proxy_ssl_server_name    on;
+        }}
+
+        location /api/lfs/ {{
+            proxy_pass               https://$upstream_hostname$request_uri;
+            proxy_set_header         authorization "";
+            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
+            proxy_set_header         job-uuid "{self._job_uuid}";
+            proxy_set_header         cookie "";
+            proxy_ssl_server_name    on;
+        }}
+
+        location /api/app/ {{
+            proxy_pass               https://$upstream_hostname$request_uri;
+            proxy_set_header         authorization "";
+            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
+            proxy_set_header         job-uuid "{self._job_uuid}";
+            proxy_set_header         cookie "";
+            proxy_ssl_server_name    on;
+        }}
+
+        location /api/resources/data-records/ {{
+            proxy_pass               https://$upstream_hostname$request_uri;
+            proxy_set_header         authorization "";
+            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
+            proxy_set_header         job-uuid "{self._job_uuid}";
+            proxy_set_header         cookie "";
+            proxy_ssl_server_name    on;
+        }}
+
+        location /api/resources/versions/ {{
+            proxy_pass               https://$upstream_hostname$request_uri;
+            proxy_set_header         authorization "";
+            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
+            proxy_set_header         job-uuid "{self._job_uuid}";
+            proxy_set_header         cookie "";
+            proxy_ssl_server_name    on;
+        }}
+
+        location /api/proxy/files/data-record-versions/ {{
+            proxy_pass               https://$upstream_hostname$request_uri;
+            proxy_set_header         authorization "";
+            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
+            proxy_set_header         job-uuid "{self._job_uuid}";
+            proxy_set_header         cookie "";
+            proxy_ssl_server_name    on;
+        }}
+
         location /api/ {{
-            proxy_pass               {base_url}/api/;
-            proxy_set_header         authorization "";
-            proxy_set_header         cookie "";
-            proxy_ssl_server_name    on;
-        }}
-
-        location /cloud-proxy/ {{
-            proxy_pass               {cloud_base_url}/cloud-proxy/;
-            proxy_set_header         authorization "";
-            proxy_set_header         cookie "";
-            proxy_ssl_server_name    on;
-        }}
-
-        location /job-storage/ {{
-            proxy_pass               {s3_results_base_url}/job-storage/;
+            proxy_pass               https://$upstream_hostname$request_uri;
             proxy_set_header         authorization "";
             proxy_set_header         cookie "";
             proxy_ssl_server_name    on;
         }}
 
         location /proxy/storage/job-storage/ {{
-            proxy_pass               {cloud_base_url}/proxy/storage/job-storage/;
+            proxy_pass               https://$upstream_hostname$request_uri;
+            proxy_set_header         authorization "";
+            proxy_set_header         cookie "";
+            proxy_ssl_server_name    on;
+        }}
+
+        location /proxy/storage/lfs/versions/ {{
+            proxy_pass               https://$upstream_hostname$request_uri;
             proxy_set_header         authorization "";
             proxy_set_header         cookie "";
             proxy_ssl_server_name    on;
         }}
 
         location /proxy/cloud/ {{
-            proxy_pass               {cloud_base_url}/proxy/cloud/;
+            proxy_pass               https://$upstream_hostname$request_uri;
             proxy_set_header         authorization "";
             proxy_set_header         cookie "";
             proxy_ssl_server_name    on;
         }}
+
+        location /proxy/index/ {{
+            proxy_pass               https://$upstream_hostname$request_uri;
+            proxy_set_header         authorization "Basic {biolib_index_basic_auth_base64}";
+            proxy_set_header         cookie "";
+            proxy_ssl_server_name    on;
+        }}
+
+        location / {{
+            return 404 "Not found";
+        }}
     }}
 }}
-'''
+"""
         else:
-            nginx_config = '''
+            nginx_config = """
 events {}
 error_log /dev/stdout info;
 stream {
-    resolver 127.0.0.11 valid=30s;'''
+    resolver 127.0.0.11 valid=30s;"""
             for idx, upstream_server_port in enumerate(upstream_server_ports):
-                nginx_config += f'''
+                nginx_config += f"""
     map "" $upstream_{idx} {{
         default {upstream_server_name}:{upstream_server_port};
     }}
@@ -327,11 +379,11 @@ stream {
     server {{
         listen          {self._remote_host['ports'][idx]} udp;
         proxy_pass      $upstream_{idx};
-    }}'''
+    }}"""
 
-            nginx_config += '''
+            nginx_config += """
 }
-'''
+"""
 
         nginx_config_bytes = nginx_config.encode()
         tarfile_in_memory = io.BytesIO()
