@@ -4,26 +4,26 @@ import json
 import os
 import time
 import tempfile
-import shutil
 import logging
-from flask import Flask, request, Response, jsonify, after_this_request
-from flask_cors import CORS  # type: ignore
+from flask import Flask, request, Response, jsonify
 
 from biolib import utils
 from biolib.biolib_api_client import BiolibApiClient
 from biolib.biolib_binary_format import SavedJob
-from biolib.compute_node.job_worker.job_storage import JobStorage
 from biolib.compute_node.webserver import webserver_utils
 from biolib.compute_node.cloud_utils.cloud_utils import CloudUtils
 from biolib.compute_node.webserver.gunicorn_flask_application import GunicornFlaskApplication
-from biolib.biolib_logging import logger, TRACE, logger_no_user_data
+from biolib.biolib_logging import logger, TRACE
 from biolib.compute_node.webserver.webserver_utils import get_job_compute_state_or_404
 from biolib.typing_utils import Optional
 
 app = Flask(__name__)
-CORS(app)
 
-biolib_tmp_dir = tempfile.TemporaryDirectory()
+if utils.IS_RUNNING_IN_CLOUD:
+    _BIOLIB_TMP_DIR = '/biolib/tmp'
+    os.makedirs(_BIOLIB_TMP_DIR, exist_ok=True)
+else:
+    _BIOLIB_TMP_DIR = tempfile.TemporaryDirectory().name
 
 
 @app.route('/hello/')
@@ -45,7 +45,7 @@ def save_job():
         return jsonify({'job': 'Invalid job'}), 400
 
     job_id = saved_job['job']['public_id']
-    job_temporary_dir = os.path.join(biolib_tmp_dir.name, job_id)
+    job_temporary_dir = os.path.join(_BIOLIB_TMP_DIR, job_id)
     os.makedirs(job_temporary_dir)
     saved_job['BASE_URL'] = BiolibApiClient.get().base_url
     saved_job['job_temporary_dir'] = job_temporary_dir
@@ -121,28 +121,8 @@ def status(job_id):
 
 
 @app.route('/v1/job/<job_id>/result/')
-def result(job_id):
-    compute_state = webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]
-    result_is_ready = compute_state['progress'] == 95
-    if result_is_ready:
-        logger_no_user_data.debug(f'Job "{job_id}" result is being downloaded by the user...')
-        job_temporary_dir = compute_state["job_temporary_dir"]
-        module_output_path = os.path.join(job_temporary_dir, JobStorage.legacy_module_output_file_name)
-        # Set the progress to 99 to inform other cleanup code that result is being downloaded
-        compute_state['progress'] = 99
-
-        @after_this_request
-        def remove_job_temporary_dir(response):  # pylint: disable=unused-variable
-            if os.path.exists(job_temporary_dir):
-                logger_no_user_data.debug(f'Removing result dir after download from path: {job_temporary_dir}')
-                shutil.rmtree(job_temporary_dir)
-            return response
-
-        return Response(open(module_output_path, 'rb'))
-
-    else:
-        logger_no_user_data.error(f'Webserver: Result for job {job_id} was not found returning 404')
-        return '', 404
+def result():
+    return 410, 'This endpoint is no longer available. Please update pybiolib by running `pip3 install -U pybiolib`'
 
 
 def send_package_to_compute_process(job_id, package_bytes):
